package fi.aalto.client;

import com.prosysopc.ua.client.UaClient;
import com.prosysopc.ua.stack.transport.security.KeyPair;
import com.prosysopc.ua.stack.transport.security.SecurityMode;
import com.prosysopc.ua.stack.core.Identifiers;
import com.prosysopc.ua.stack.builtintypes.DataValue;
import com.prosysopc.ua.stack.builtintypes.DateTime;
import com.prosysopc.ua.stack.builtintypes.LocalizedText;
import com.prosysopc.ua.stack.builtintypes.NodeId;
import com.prosysopc.ua.stack.builtintypes.Variant;
import com.prosysopc.ua.stack.transport.security.AllowAllCertificatesValidator;
import com.prosysopc.ua.stack.cert.PkiDirectoryCertificateStore;
import com.prosysopc.ua.stack.core.ApplicationDescription;
import com.prosysopc.ua.stack.core.Attributes;
import com.prosysopc.ua.stack.core.MonitoringMode;
import com.prosysopc.ua.nodes.UaNode;
import com.prosysopc.ua.nodes.UaReference;
import com.prosysopc.ua.UaAddress;
import com.prosysopc.ua.UaApplication;
import com.prosysopc.ua.UserIdentity;
import com.prosysopc.ua.ApplicationIdentity;
import com.prosysopc.ua.client.MonitoredDataItem;
import com.prosysopc.ua.client.Subscription;

import java.io.File;
import java.util.Locale;

import fi.aalto.client.SimpleMonitoredDataItemListener;

public class SimpleClient {
	
	protected static String APP_NAME = "SimpleClient";
	protected UaAddress serverAddress;
	UaClient client = new UaClient();
	Subscription subscription;
	
	public void run() { 	

		try {
			// Setup security
		    ApplicationDescription appDescription = new ApplicationDescription();
		    appDescription.setApplicationName(new LocalizedText(APP_NAME, Locale.ENGLISH));
		    appDescription.setApplicationUri("urn:localhost:OPCUA" + APP_NAME);
			appDescription.setProductUri("urn:prosysopc.com:OPCUA" + APP_NAME);

		    final PkiDirectoryCertificateStore certificateStore = new PkiDirectoryCertificateStore();
		    final AllowAllCertificatesValidator validator = new AllowAllCertificatesValidator();
			client.setCertificateValidator(validator);
			File privatePath = new File(certificateStore.getBaseDir(), "private");
			KeyPair issuerCertificate = ApplicationIdentity.loadOrCreateIssuerCertificate(
					"ClientCA", privatePath, "client", 3650, false);
			int[] keySizes = new int[] { 0, 4096 };
		    final ApplicationIdentity identity = ApplicationIdentity
					.loadOrCreateCertificate(appDescription, "Client Organisation",
					"client", privatePath, issuerCertificate, keySizes, true);
		    client.setApplicationIdentity(identity);
			
		    // Task 1. Connect to a server
		    this.serverAddress = UaAddress.fromComponents(UaApplication.Protocol.OpcTcp, "localhost", 52500, "DemoServer");
			this.client.setAddress(serverAddress); 
	        this.client.setSecurityMode(SecurityMode.BASIC256SHA256_SIGN_ENCRYPT);
	        this.client.setUserIdentity(new UserIdentity("opcua", "opcua"));
	        this.client. // TODO 

		    System.out.println("Connected to " + this.serverAddress.toString());
	        
		    // Task 2. Find some nodes
		    // RootFolder
		    UaNode node = // TODO find a node identified as Identifiers.RootFolder
		    
		    System.out.println("Name of RootFolder is " + node.getDisplayName().toString());
		    
		    // get child nodes of RootFolder
		    System.out.println("Children of Root are:");
		    UaReference[] references = node. // TODO find downward references of type Identifiers.Organizes
		    for (int i = 0; i < references.length; i++) {
		    	UaNode oppositeNode = references[i].getOppositeNode(node);
			    System.out.println(i+1 + ". " +oppositeNode.getDisplayName().toString());		    	
		    }
		    
		    // Task 3. Write value to a variable
		    Variant variantValue = new Variant(new Float(1.25));
		    this.client. // TODO write to NodeId(2, "P300_MeasMan")
		    
		    System.out.println("New value should be " + variantValue.floatValue());
			
		    // Task 4. Read values from variables		
			DataValue value = this.client.readValue(Identifiers.Server_ServerStatus_CurrentTime);
			DateTime dateTimeValue = // TODO get DateTime from DataValue
			value = this.client.readValue(new NodeId(2, "P300_MeasMan")); 
			float floatValue = // TODO get floatValue from DataValue

		    System.out.println("Value is " + floatValue);
		    System.out.println("Time is " + dateTimeValue.toString());

		    // Task 5. Subscribe to a variable
		    MonitoredDataItem monitoredItem = new MonitoredDataItem(new NodeId(2, "P300_MeasVal"), Attributes.Value, MonitoringMode.Reporting);
		    SimpleMonitoredDataItemListener dataChangeListener = new SimpleMonitoredDataItemListener(this);
		    monitoredItem.setDataChangeListener(dataChangeListener);
			this.subscription = new Subscription();
			this.subscription. // TODO add the monitored item to the subscription
			dataChangeListener.setSubscription(this.subscription);
			this.client. // TODO add the subscription to the client
			
		    System.out.println("Start subscription");
		    Thread.sleep(10000);

		    // Task 6. Unsubscribe from a variable
		    this.client. // TODO

		    System.out.println("End subscription");

		    // Task 7. Disconnect from the server
	        this.client. // TODO
	        
		    System.out.println("Disconnected from " + this.serverAddress.toString());
	    
		} catch  (Exception e) {
			System.out.println(e.getMessage());
			return;
		}
		return;
	}
	
	public static void main(String[] args) {
		
		new SimpleClient().run();
	}

}

